/*
 * naalaa_image.c
 * --------------
 *
 * By: Marcus 2021
 */

#include <stdlib.h>
#include <stdio.h> 
#include "stb_image.h"
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include "stb_image_write.h"
#include "naalaa_image.h"

void IMG_UpdateAlphaInfo(Image *img);
void IMG_DrawHLine(Image *dst, int xstart, int xend, int row, unsigned int color);

/*
 * IMG_Create
 * ----------
 */
Image *IMG_Create(int w, int h, unsigned int color) {
    Image *img;
    int i;
    
    if (w <= 0 || h <= 0) return 0;
    
    img = (Image *)malloc(sizeof(Image));
    img->buffer = (unsigned int *)malloc(sizeof(unsigned int)*w*h);
    img->cellInfo = (ImageCellInfo *)malloc(sizeof(ImageCellInfo));
    img->w = w;
    img->h = h;
    img->hasAlpha = ColorAlphaComponent(color) < 128;
    img->cellInfo[0].hasAlpha = img->hasAlpha;
    img->hasColorKey = 0;
    img->colorKey = 0x80ff00ff;
    img->cols = 1;
    img->rows = 1;
    img->cells = 1;
    img->xMin = 0;
    img->yMin = 0;
    img->xMax = w;
    img->yMax = h;
    
    for (i = 0; i < w*h; i++) img->buffer[i] = color;
    
    return img;
}

/*
 * IMG_FromBuffer
 * --------------
 */
Image *IMG_FromBuffer(unsigned int *buffer, int w, int h) {
    Image *img;
    
    if (!buffer || w <= 0 || h <= 0) return 0;
    
    img = (Image *)malloc(sizeof(Image));
    img->buffer = buffer;
    img->cellInfo = (ImageCellInfo *)malloc(sizeof(ImageCellInfo));
    img->w = w;
    img->h = h;
    img->hasAlpha = 0;
    img->hasColorKey = 0;
    img->colorKey = 0x80ff00ff;
    img->cols = 1;
    img->rows = 1;
    img->cells = 1;
    img->xMin = 0;
    img->yMin = 0;
    img->xMax = w;
    img->yMax = h;
    
    return img;
}

/*
 * IMG_Load
 * --------
 */
Image *IMG_Load(const char *filename) {
    Image *img;
    unsigned int *buffer;
    int w, h;
    
    int comp;
    buffer = (unsigned int *)stbi_load(filename, &w, &h, &comp, 4);
    if (buffer) {
        int s = w*h;
        
        img = IMG_FromBuffer(buffer, w, h);
        for (int i = 0; i < s; i++) {
            unsigned char r, g, b, a;
            r = (buffer[i] & 0xff);
            g = (buffer[i] & 0xff00) >> 8;
            b = (buffer[i] & 0xff0000) >> 16;
            a = (buffer[i] & 0xff000000) >> 24;
            if (a == 255) {
                a = 128;
            }
            else {
                img->hasAlpha = 1;
                a /= 2;
            }
            buffer[i] = ToRGBA(r, g, b, a);
        }
        img->cellInfo[0].hasAlpha = img->hasAlpha;
        return img;
    }
    else {
        return 0;
    }
}

/*
 * IMG_Save
 * --------
 */
int IMG_Save(Image *img, const char *filename) {
	unsigned int *pngBuffer = 0;
    
	pngBuffer = (unsigned int *)malloc(sizeof(unsigned int)*img->w*img->h);
	for (int i = 0; i < img->w*img->h; i++) {
		unsigned char r, g, b, a;
		r = (img->buffer[i] & 0xff0000) >> 16;
		g = (img->buffer[i] & 0xff00) >> 8;
		b = (img->buffer[i] & 0xff);
		a = (img->buffer[i] & 0xff000000) >> 24;
        if (a == 128) a = 255;
        else a *= 2;
		pngBuffer[i] = (r) + (g << 8) + (b << 16) + (a << 24);
	}
	int result = stbi_write_png(filename, img->w, img->h, 4, pngBuffer, img->w*4);
	free(pngBuffer);
    
	return result != 0;
}

/*
 * IMG_Free
 * --------
 */
void IMG_Free(Image *img) {
    free(img->buffer);
    free(img->cellInfo);
    free(img);
}

/*
 * IMG_SetColorKey
 * ---------------
 */
void IMG_SetColorKey(Image *img, unsigned int color) {
    img->hasColorKey = 1;
    img->colorKey = color;
    IMG_BufferChanged(img);
}

/*
 * IMG_SetGrid
 * -----------
 */
void IMG_SetGrid(Image *img, int cols, int rows) {
    if (img->w%cols > 0 || img->h%rows > 0) return;
    
    img->cols = cols <= 0 ? cols = 1 : cols;
    img->rows = rows <= 0 ? rows = 1 : rows;
    img->cells = img->cols*img->rows;
    free(img->cellInfo);
    img->cellInfo = (ImageCellInfo *)malloc(sizeof(ImageCellInfo)*img->cells);
    IMG_UpdateAlphaInfo(img);
}


/*
 * IMG_Width
 * ---------
 */
int IMG_Width(Image *img) {
    return img->w;
}

/*
 * IMG_Height
 * ----------
 */
int IMG_Height(Image *img) {
    return img->h;
}

/*
 * IMG_Buffer
 * ----------
 */
unsigned int *IMG_Buffer(Image *img) {
    return img->buffer;
}

/*
 * IMG_SetClipRect
 * ---------------
 */
void IMG_SetClipRect(Image *img, int x, int y, int w, int h) {
    img->xMin = x;
    img->yMin = y;
    img->xMax = x + w;
    img->yMax = y + h;

    if (img->xMin < 0) img->xMin = 0;
    if (img->xMin >= img->w) img->xMin = img->w - 1;
    if (img->yMin < 0) img->yMin = 0;
    if (img->yMin >= img->h) img->yMin = img->h - 1;

    if (img->xMax < 0) img->xMax = 0;
    if (img->xMax > img->w) img->xMax = img->w;
    if (img->yMax < 0) img->yMax = 0;
    if (img->yMax > img->h) img->yMax = img->h;
}

/*
 * IMG_ClipRect
 * ------------
 */
void IMG_ClipRect(Image *img, int x, int y, int w, int h) {
}

/*
 * IMG_ClearClipRect
 * -----------------
 */
void IMG_ClearClipRect(Image *img) {
    img->xMin = 0;
    img->yMin = 0;
    img->xMax = img->w;
    img->yMax = img->h;
}

/*
 * IMG_ClipX
 * ---------
 */
int IMG_ClipX(Image *img) {
    return img->xMin;
}

/*
 * IMG_ClipY
 * ---------
 */
int IMG_ClipY(Image *img) {
    return img->yMin;
}

/*
 * IMG_ClipWidth
 * -------------
 */
int IMG_ClipWidth(Image *img) {
    return img->xMax - img->xMin;
}

/*
 * IMG_ClipHeight
 * --------------
 */
int IMG_ClipHeight(Image *img) {
    return img->yMax - img->yMin;
}

/*
 * IMG_DrawImage
 * -------------
 */
void IMG_DrawImage(Image *dst, int x, int y, Image *src, int srcX, int srcY, int srcW, int srcH, unsigned int color, int useImageAlpha) {
    unsigned int *srcBuffer;
    unsigned int *dstBuffer;
    int width, height;
    int sWidth;
    unsigned char srcR, srcG, srcB, srcA, dstR, dstG, dstB, invA;
    int r, g, b, a;
    unsigned char cR, cG, cB, cA;
    int hasAlpha, hasColor;
    int i, j;

    if (srcW <= 0 || srcH <= 0) return;
    if (srcX < 0) {
        srcW += srcX;
        srcX = 0;
        if (srcW <= 0) return;
    }
    if (srcY < 0) {
        srcH += srcY;
        srcY = 0;
        if (srcH <= 0) return;
    }
    if (srcX + srcW > src->w) srcW -= srcX + srcW - src->w;
    if (srcY + srcH > src->h) srcH -= srcY + srcH - src->h;
    if (x + srcW < dst->xMin || x > dst->xMax) return;
    if (y + srcH < dst->yMin || y > dst->yMax) return;
    
    srcBuffer = src->buffer + src->w*srcY + srcX;
    dstBuffer = dst->buffer;
        
    width = srcW;
    height = srcH;

    if (x < dst->xMin) {
        srcBuffer += dst->xMin - x;
        width -= dst->xMin - x;
        dstBuffer += dst->xMin;
    }
    else dstBuffer += x;

    if (y < dst->yMin) {
        srcBuffer += (dst->yMin - y)*src->w;
        height -= dst->yMin - y;
        dstBuffer += dst->yMin*dst->w;
    }
    else dstBuffer += y*dst->w;

    if (x + srcW > dst->xMax) width -= x + srcW - dst->xMax;

    if (y + srcH > dst->yMax) height -= y + srcH - dst->yMax;
        
    ColorToRGBAComponents(color, cR, cG, cB, cA);
    if (cA == 0) return;

    sWidth = src->w;

    hasColor = cR != 255 || cG != 255 || cB != 255;
    hasAlpha = cA != 128 || (src->hasAlpha && useImageAlpha);

    if (hasColor && hasAlpha) {
        for (i = 0; i < height; i++, srcBuffer += sWidth, dstBuffer += dst->w) {
            for (j = 0; j < width; j++) {
                ColorToRGBComponents(dstBuffer[j], dstR, dstG, dstB);
                ColorToRGBAComponents(srcBuffer[j], srcR, srcG, srcB, srcA);
                if (srcA != 0) {
                    r = (srcR * cR) >> 8;
                    g = (srcG * cG) >> 8;
                    b = (srcB * cB) >> 8;
                    a = (srcA * cA) >> 7;
                    invA = 128 - a;
                    dstR = (dstR*invA + r*a) >> 7;
                    dstG = (dstG*invA + g*a) >> 7;
                    dstB = (dstB*invA + b*a) >> 7;
                    dstBuffer[j] = ToRGB(dstR, dstG, dstB);
                }
            }
        }
    }
    else if (hasAlpha) {
        for (i = 0; i < height; i++, srcBuffer += sWidth, dstBuffer += dst->w) {
            for (j = 0; j < width; j++) {
                ColorToRGBComponents(dstBuffer[j], dstR, dstG, dstB);
                ColorToRGBAComponents(srcBuffer[j], srcR, srcG, srcB, srcA);
                if (srcA != 0) {
                    a = (srcA * cA) >> 7;
                    invA = 128 - a;
                    dstR = (dstR*invA + srcR*a) >> 7;
                    dstG = (dstG*invA + srcG*a) >> 7;
                    dstB = (dstB*invA + srcB*a) >> 7;
                    dstBuffer[j] = ToRGB(dstR, dstG, dstB);
                }
            }
        }
    }
    else if (hasColor) {
        for (i = 0; i < height; i++, srcBuffer += sWidth, dstBuffer += dst->w) {
            for (j = 0; j < width; j++) {
                ColorToRGBComponents(srcBuffer[j], srcR, srcG, srcB);
                r = (srcR * cR) >> 8;
                g = (srcG * cG) >> 8;
                b = (srcB * cB) >> 8;
                dstBuffer[j] = ToRGB(r, g, b);
            }
        }
    }
    else {
        for (i = 0; i < height; i++, srcBuffer += sWidth, dstBuffer += dst->w) {
            for (j = 0; j < width; j++) {
                dstBuffer[j] = srcBuffer[j];
            }
        }
    }
}

/*
 * IMG_DrawImageCel
 * ----------------
 */
void IMG_DrawImageCel(Image *dst, int x, int y, Image *src, int cel, unsigned int color) {
    int celW = src->w/src->cols;
    int celH = src->h/src->rows;
    if (cel >= 0 || cel < src->cells) IMG_DrawImage(dst, x, y, src, celW*(cel%src->cols), celH*(cel/src->cols), celW, celH, color, src->cellInfo[cel].hasAlpha);
}


/*
 * IMG_DrawPixel
 * -------------
 */
void IMG_DrawPixel(Image *dst, int x, int y, unsigned int color) {
    unsigned char srcR, srcG, srcB, srcA, dstR, dstG, dstB, invA;
    int ai;
    
    if (x < dst->xMin || x >= dst->xMax || y < dst->yMin || y >= dst->yMax) return;
        
    ai = y*dst->w + x;
    if (ColorAlphaComponent(color) < 128) {
        ColorToRGBAComponents(color, srcR, srcG, srcB, srcA);
        ColorToRGBComponents(dst->buffer[ai], dstR, dstG, dstB);
        invA = 128 - srcA;
        dstR = (dstR*invA + srcR*srcA) >> 7;
        dstG = (dstG*invA + srcG*srcA) >> 7;
        dstB = (dstB*invA + srcB*srcA) >> 7;
        
        color = ToRGB(dstR, dstG, dstB);
    }
    
    dst->buffer[ai] = color;
}

/*
 * IMG_SetPixel
 * ------------
 */
void IMG_SetPixel(Image *dst, int x, int y, unsigned int color) {
    if (x < dst->xMin || x >= dst->xMax || y < dst->yMin || y >= dst->yMax) return;
        
    dst->buffer[y*dst->w + x] = color;
}

/*
 * IMG_GetPixel
 * ------------
 */
int IMG_GetPixel(Image *img, int x, int y, unsigned int *color) {
    if (x < img->xMin || x >= img->xMax || y < img->yMin || y >= img->yMax) return 0;
    *color = img->buffer[y*img->w + x];
    return 1;
}


/*
 * IMG_DrawRect
 * ------------
 */
void IMG_DrawRect(Image *dst, int x, int y, int w, int h, unsigned int color) {
    IMG_FillRect(dst, x, y, w, 1, color);
    IMG_FillRect(dst, x, y + h - 1, w, 1, color);
    IMG_FillRect(dst, x, y + 1, 1, h - 1, color);
    IMG_FillRect(dst, x + w - 1, y + 1, 1, h - 1, color);
}

/*
 * IMG_FillRect
 * ------------
 */
void IMG_FillRect(Image *dst, int x, int y, int w, int h, unsigned int color) {
    unsigned int *cb;
    int dx;
    int x1, y1;

    if (w <= 0 || h <= 0) return;

    x1 = x + w;
    y1 = y + h;
        
    if (x >= dst->xMax || y >= dst->yMax || x1 < dst->xMin || y1 < dst->yMin) return;
                
    if (x < dst->xMin) x = dst->xMin;
    if (x1 > dst->xMax) x1 = dst->xMax;
    if (y < dst->yMin) y = dst->yMin;
    if (y1 > dst->yMax) y1 = dst->yMax;
        
    cb = dst->buffer + y*dst->w + x;
    dx = x1 - x;
        
    if (ColorAlphaComponent(color) == 128) {
        for (int ys = y; ys < y1; ys++, cb += dst->w) {
            for (int xs = 0; xs < dx; xs++) {
                cb[xs] = color;
            }
        }
    }
    else {
        unsigned char srcR, srcG, srcB;
        unsigned char srcA, dstR, dstG, dstB, invA;
        
        ColorToRGBAComponents(color, srcR, srcG, srcB, srcA);
        srcR = (srcR*srcA) >> 7;
        srcG = (srcG*srcA) >> 7;
        srcB = (srcB*srcA) >> 7;
        invA = 128 - srcA;
        for (int ys = y; ys < y1; ys++, cb += dst->w) {
            for (int xs = 0; xs < dx; xs++) {
                ColorToRGBComponents(cb[xs], dstR, dstG, dstB);
                dstR = ((dstR*invA) >> 7) + srcR;
                dstG = ((dstG*invA) >> 7) + srcG;
                dstB = ((dstB*invA) >> 7) + srcB;
                cb[xs] = ToRGBA(dstR, dstG, dstB, 128);                            
            }
        }
    }
}

/*
 * IMG_ClearRect
 * -------------
 */
void IMG_ClearRect(Image *dst, int x, int y, int w, int h) {
    unsigned int *cb;
    int dx;
    int x1, y1;

    if (w <= 0 || h <= 0) return;

    x1 = x + w;
    y1 = y + h;
        
    if (x >= dst->xMax || y >= dst->yMax || x1 < dst->xMin || y1 < dst->yMin) return;
                
    if (x < dst->xMin) x = dst->xMin;
    if (x1 > dst->xMax) x1 = dst->xMax;
    if (y < dst->yMin) y = dst->yMin;
    if (y1 > dst->yMax) y1 = dst->yMax;
        
    cb = dst->buffer + y*dst->w + x;
    dx = x1 - x;
        
    for (int ys = y; ys < y1; ys++, cb += dst->w) {
        for (int xs = 0; xs < dx; xs++) {
            cb[xs] = 0x00000000;
        }
    }
    dst->hasAlpha = 1;
}

/*
 * IMG_DrawLine
 * ------------
 * Lines should be clipped and an unsafe version of IMG_DrawPixel should be used.
 */
void IMG_DrawLine(Image *dst, int x0, int y0, int x1, int y1, unsigned int color) {
    int dx, sx, dy, sy, err;
    
    if (y0 == y1) {
        if (x1 >= x0) IMG_FillRect(dst, x0, y0, x1 - x0 + 1, 1, color);
        else  IMG_FillRect(dst, x1, y0, x0 - x1 + 1, 1, color);
        return;
    }
    else if (x0 == x1) {
        if (y1 >= y0) IMG_FillRect(dst, x0, y0, 1, y1 - y0 + 1, color);
        else  IMG_FillRect(dst, x0, y1, 1, y0 - y1 + 1, color);
        return;
    }
    
    dx = x1 > x0 ? x1 - x0 : x0 - x1;
    sx = x0 < x1 ? 1 : -1;
    dy = y1 > y0 ? y0 - y1 : y1 - y0;
    sy = y0 < y1 ? 1 : -1;
    err = dx + dy;
    while (1) {
        int e2;
        IMG_DrawPixel(dst, x0, y0, color);
        if (x0 == x1 && y0 == y1) break;
        e2 = 2*err;
        if (e2 >= dy) {
            err += dy;
            x0 += sx;
        }
        if (e2 <= dx) {
            err += dx;
            y0 += sy;
        }
    }
}

/*
 * IMG_FillPolygon
 * ---------------
 */
void IMG_FillPolygon(Image *dst, int numPoints, int *points, unsigned int color) {
    int minY, maxY, minX, maxX;
    int i, y;
    int *p;
    int xval[16];
    int numxval;
    
    if (numPoints <= 0) return;

    p = points;
    minY = *(p + 1);
    maxY = minY;
    minX = *p;
    maxX = minX;
    p += 2;
    for (i = 1; i < numPoints; i++, p += 2) {
        if (*(p + 1) < minY) minY = *(p + 1);
        if (*(p + 1) > maxY) maxY = *(p + 1);
        if (*p < minX) minX = *p;
        if (*p > maxX) maxX = *p;
    }
    if (maxX < dst->xMin || minX >= dst->xMax || maxY < dst->yMin || minY >= dst->yMax) return;

    for (y = minY; y <= maxY; y++) {
        numxval = 0;
        for (i = 0; i < numPoints; i++) {
            int dx, dy;
            int *p0 = &points[i*2];
            int *p1 = &points[((i + 1)%numPoints)*2];
            if ((p0[1] <= y && p1[1] > y) || (p0[1] > y && p1[1] <= y)) {
                dy = p1[1] - p0[1];
                if (dy == 0) {
                    xval[numxval++] = p0[0];
                    xval[numxval++] = p1[0];
                }
                else {
                    dx = p1[0] - p0[0];
                    dy = p1[1] - p0[1];
                    xval[numxval++] = (p0[0] + dx*(y - p0[1])/dy);
                }
            }
        }
        if (numxval) {
            char sorted;
            int last = numxval - 1;
            do {
                sorted = 1;
                for (i = 0; i < last; i++) {
                    if (xval[i] > xval[i + 1]) {
                        sorted = 0;
                        int tmp = xval[i];
                        xval[i] = xval[i + 1];
                        xval[i + 1] = tmp;
                    }
                }
                last--;
            } while (!sorted);
            for (i = 0; i < numxval; i += 2) {
                IMG_DrawHLine(dst, xval[i], xval[i + 1], y, color);
            }
        }
    }
}

void IMG_DrawPolygon(Image *dst, int numPoints, int *points, unsigned int color) {
    if (numPoints <= 0) return;
    
    for (int i = 0; i < numPoints; i++) {
        int *p0 = &points[i*2];
        int *p1 = &points[((i + 1)%numPoints)*2];
        IMG_DrawLine(dst, p0[0], p0[1], p1[0], p1[1], color);
    }
}

/*
 * IMG_DrawHLine
 * -------------
 */
void IMG_DrawHLine(Image *dst, int xstart, int xend, int row, unsigned int color) {
    int w;
    unsigned char dstR, dstG, dstB, invA;
    unsigned char cR, cG, cB, cA;
    unsigned int *cb;
        
    if (row < dst->yMin || row >= dst->yMax) return;
    
    w = xend - xstart;
    if (w <= 0) return;
    if (xstart >= dst->xMax) return;
    if (xend < dst->xMin) return;
    
    if (xstart < dst->xMin) xstart = dst->xMin;
    if (xend >= dst->xMax) xend = dst->xMax - 1;

    ColorToRGBAComponents(color, cR, cG, cB, cA);
    cb = &dst->buffer[row*dst->w + xstart];

    if (cA < 128) {
        invA = 128 - cA;
        for (int i = xstart; i <= xend; i++, cb++) {
            ColorToRGBComponents(*cb, dstR, dstG, dstB);
            dstR = (dstR*invA + cR*cA) >> 7;
            dstG = (dstG*invA + cG*cA) >> 7;
            dstB = (dstB*invA + cB*cA) >> 7;
            *cb = ToRGB(dstR, dstG, dstB);
        }
    }
    else {
        for (int i = xstart; i <= xend; i++, cb++) *cb = color;
    }
}

void IMG_UpdateAlphaInfo(Image *img) {
    int rowh = img->h/img->rows;
    int colw = img->w/img->cols;

    for (int row = 0; row < img->rows; row++) {
        for (int col = 0; col < img->cols; col++) {
            unsigned int *cb = img->buffer + row*rowh*img->w + col*colw;
            char hasAlpha = 0;
            for (int y = 0; y < rowh && !hasAlpha; y++, cb += img->w) {
                for (int x = 0; x < colw && !hasAlpha; x++) {
                    hasAlpha = ColorAlphaComponent(cb[x]) < 128;
                }
            }
            if (hasAlpha) {
                img->hasAlpha = 1;
                img->cellInfo[row*img->cols + col].hasAlpha = 1;
            }
            else {
                img->cellInfo[row*img->cols + col].hasAlpha = 0;
            }
        }
    }
}

void IMG_BufferChanged(Image *img) {
    if (img->hasColorKey) {
        for (int i = 0; i < img->w*img->h; i++) {
            if (img->buffer[i] == img->colorKey) img->buffer[i] = 0x0;
        }
    }
    IMG_UpdateAlphaInfo(img);
}
